#ifndef _BROKER_H
#define _BROKER_H

#include <linux/kobject.h>
#include <linux/list.h>

#define TOPIC_MAX_NAME_LEN 32
#define TOPIC_MAX_VALUE_LEN 128

/* Common publisher and susbscription options */
#define TOPIC_DEFAULTS 0x0000           /* no creation/subscription options */
#define TOPIC_SYSFS_READ 0x0001         /* expose the topic over sysfs for read access */
#define TOPIC_SYSFS_WRITE 0x0002        /* expose the topic over sysfs for write access */
#define TOPIC_SYSFS_READ_WRITE (TOPIC_SYSFS_READ | TOPIC_SYSFS_WRITE)

#define ssf_debug(...) pr_debug("[" SSF_MODULE_NAME "] " __VA_ARGS__)
#define ssf_info(...) pr_info("[" SSF_MODULE_NAME "] " __VA_ARGS__)
#define ssf_warn(...) pr_warn("[" SSF_MODULE_NAME "] " __VA_ARGS__)
#define ssf_err(...) pr_err("[" SSF_MODULE_NAME "] " __VA_ARGS__)
#define ssf_devel(...) pr_devel("[" SSF_MODULE_NAME "] " __VA_ARGS__)

struct topic_t;

/**
 * int check_bounds_fn_t(topic, new_value, old_value, source_name)
 * Callback used for keeping the topic value in specified bounds.
 * @topic: Pointer to the topic that has changed its value.
 * @new_value: The new value (current) of the topic.
 * @old_value: The old value of the topic prior to the change.
 * @source_name: The name of the kernel module or subsystem that has changed
 *   the topic's value.
 * @data: Payload data which can be provided when subscribing
 * @return: 0 if successful (input in bounds).
 */
typedef int (*check_bounds_fn_t)(struct topic_t *topic, const char *new_value,
				  const char *old_value, const char *source_name, void *data);

/**
 * void set_bounds_checker_of_topic(const char *topic_name, check_bounds_fn_t fn)
 * Set a callback function to be used for keeping the topic value in specified bounds.
 * @topic_name: Name of the topic.
 * @fn: The function where you can check the bounds.
 */
void set_bounds_checker_of_topic(const char *topic_name, check_bounds_fn_t fn);

/**
 * struct topic_t - A topic combines a string value with a notification
 * mechanism for when the topic's value is changed. Think of it as a
 * publish/subscribe mechanism.
 *
 * @list_node: Used internally to keep a list of registered topics.
 * @name: The topic's name. Will also be used as sysfs filename.
 * @value: The topic's current value.
 * @options: Whether the topic shall be exported to user space via sysfs.
 *   Pass TOPIC_SYSFS_READ, TOPIC_SYSFS_WRITE, TOPIC_SYSFS_READ_WRITE, or
 *   TOPIC_DEFAULTS. The latter does not expose the topic over sysfs.
 * @kobj_attr: Internally used to create a sysfs file.
 * @subscription_list: A list of all registered subscriptions, i.e. links
 *   topic_subscription_t objects.
 *
 * @note Topics are created implicily by preparing a publisher or subscribing
 * a subscription. See prepare_topic_publisher() and subscribe_to_topic().
 */
struct topic_t {
	struct list_head	list_node;
	char			name[TOPIC_MAX_NAME_LEN];
	char			value[TOPIC_MAX_VALUE_LEN];
	unsigned int		options;
	struct kobj_attribute	kobj_attr;
	struct list_head	subscription_list;
	check_bounds_fn_t check_bounds_fn;
};

/**
 * get_topic(topic_name) - Returns a topic by name.
 * @topic_name: The topic's name.
 * @returns A pointer to the topic of NULL if the topic does not exist.
 */
struct topic_t *get_topic(const char *topic_name);

/**
 * get_topic_value(topic_name, default_value) - Returns the current value
 * of a topic.
 * @topic_name: The topic's name.
 * @default_value: The value to return if the topic does not exist.
 * @returns The current value of the topic if found by name or the value of the
 * default_value parameter if the topic does not exist.
 */
const char *get_topic_value(const char *topic_name, const char *default_value);

/**
 * struct topic_publisher_t - A publish can be used to update the value of a
 * topic.
 * @topic_name: The name of the topic.
 * @source_name: Name of the kernel module or subsystem that owns this publisher
 *   and that is going to modify the topic's value.
 * @options: Whether the topic shall be exported to user space via sysfs.
 *   Pass TOPIC_SYSFS_READ, TOPIC_SYSFS_WRITE, TOPIC_SYSFS_READ_WRITE, or
 *   TOPIC_DEFAULTS. The latter does not expose the topic over sysfs.
 * @topic: The pointer to the topic is populated by prepare_topic_publisher().
 *
 * @note Use DEFINE_TOPIC_PUBLISHER() to define a topic as static variable.
 */
struct topic_publisher_t {
	const char *	topic_name;
	const char *	source_name;
	unsigned int	options;
	struct topic_t *topic;
};

/**
 * TOPIC_PUBLISHER_INIT(name, source, opts) - Initialize a publisher.
 * @name: The name of the topic to publish to.
 * @source: The name of the kernel module or subsystem that will publish.
 * @opts: Whether the topic shall be exported to user space via sysfs.
 *   Pass TOPIC_SYSFS_READ, TOPIC_SYSFS_WRITE, TOPIC_SYSFS_READ_WRITE, or
 *   TOPIC_DEFAULTS. The latter does not expose the topic over sysfs.
 *
 * @example
 * static my_pub = TOPIC_PUBLISHER_INIT("my_topic", "my_module",
 *   TOPIC_SYSFS_READ);
 */
#define TOPIC_PUBLISHER_INIT(name, source, opts)                               \
	{                                                                      \
		.topic_name = (name),                                          \
		.topic = NULL,                                                 \
		.source_name = (source),                                       \
		.options = (opts)                                              \
	}

/**
 * DEFINE_TOPIC_PUSLISHER(publisher, name, source, opts) - Defines a publisher.
 * @publisher: Publisher variable to define.
 * @name: The name of the topic to publish to.
 * @source: The name of the kernel module or subsystem that will publish.
 * @opts: Whether the topic shall be exported to user space via sysfs.
 *   Pass TOPIC_SYSFS_READ, TOPIC_SYSFS_WRITE, TOPIC_SYSFS_READ_WRITE, or
 *   TOPIC_DEFAULTS. The latter does not expose the topic over sysfs.
 *
 * @example
 * static DEFINE_TOPIC_PUBLISHER(my_pub, "my_topic", "my_module",
 *   TOPIC_SYSFS_READ);
 */
#define DEFINE_TOPIC_PUBLISHER(publisher, name, source, opts)                  \
	struct topic_publisher_t publisher =                                   \
		TOPIC_PUBLISHER_INIT(name, source, opts)

/**
 * prepare_topic_publisher(publisher, value) - Registers a publisher with a topic.
 * @publisher: topic_publisher_t struct, previously defined with
 *   DEFINE_TOPIC_PUBLISHER() or initialzed with TOPIC_PUBLSIHER_INIT().
 * @value: Optional value to be published initially or NULL.
 * @returns 0 if successful.
 *
 * This creates the topic if it does not exist already. Depending on the
 * module load order, another module could have created the topic already
 * by subscribing to it (we allow subscribe-first flows). After this step
 * publish_to_topic() calls are valid.
 *
 * Note: This function must be called while running in the nmworker thread.
 * If called from outside this worker thread, call synchronize_to_worker()
 * before calling one of the functions and unsynchronize_from_worker() after.
 */
int prepare_topic_publisher(struct topic_publisher_t *publisher, const char *value);

/**
 * publish_to_topic(publisher, value) - Updates the topic with a new value.
 * @publisher: topic_publisher_t struct, previously prepared with
 *   prepare_topic_publisher().
 * @value: New value to set.
 * @return 0 if successful.
 *
 * Updates the current value of the topic and notifies all subscribers.
 *
 * @note This function must be called while running in the nmworker thread.
 * If called from outside this worker thread, call synchronize_to_worker()
 * before calling one of the functions and unsynchronize_from_worker() after.
 */
int publish_to_topic(struct topic_publisher_t *publisher, const char *value);

/* Subscription options */
#define TOPIC_NOTIFY_NOW 0x0010         /* notify immediately with current value */
#define TOPIC_NOTIFY_SAME 0x0020        /* notify when changed to same value */

struct topic_subscription_t;

/**
 * topic_notify_fn_t(topic, new_value, old_value, source_name) - Notificaiton
 * callback used by topic subscribers.
 * @topic: Pointer to the topic that has changed its value.
 * @new_value: The new value (current) of the topic.
 * @old_value: The old value of the topic prior to the change.
 * @source_name: The name of the kernel module or subsystem that has changed
 *   the topic's value.
 * @data: Payload data which can be provided when subscribing
 */
typedef void (*topic_notify_fn_t)(struct topic_t *topic, const char *new_value,
				  const char *old_value, const char *source_name, void *data);

/**
 * struct topic_subscription_t - Structure that holds information about a
 * topic subscription.
 * @list_node: Used internally to keep a list of subscriptions per topic.
 *   Will be linked from the subscription_list field in struct topic_t.
 * @topic_name: The name of the topic.
 * @notify_call: The function to call whenever the value of the topic has
 *   changed.
 * @sink_name: Name of the kernel module or subsystem that owns this
 *   subscription.
 * @options: Whether the topic shall be exported to user space via sysfs.
 *   Pass TOPIC_SYSFS_READ, TOPIC_SYSFS_WRITE, TOPIC_SYSFS_READ_WRITE, or
 *   TOPIC_DEFAULTS. The latter does not expose the topic over sysfs.
 *   Also, supports the options TOPIC_NOTIFY_NOW to indicate that the callback
 *   shall be called immediately with the current value for snapshotting and
 *   TOPIC_NOTIFY_SAME to indicate that the callback shall be called when
 *   a publisher sets the same value again.
 * @topic: The pointer to the topic is populated by prepare_topic_publisher().
 * @data: Payload data which can be used to e.g. store device information
 * @note Use DEFINE_TOPIC_SUBSCRIPTION() or DEFINE_GLOBAL_TOPIC_SUBSCRIPTION()
 *   to define a topic as static variable.
 */
struct topic_subscription_t {
	struct list_head	list_node;
	const char *		topic_name;
	topic_notify_fn_t	notify_call;
	const char *		sink_name;
	unsigned int		options;
	struct topic_t *	topic;
	void *			data;
};

/**
 * TOPIC_SUBSCRIPTION_INIT(subscription, name, fn, sink, opts) - Initialize a
 * subscription for a particular topic referred-to by its name.
 * @subscription: Subscription variable to initialize.
 * @name: The name of the topic to subscribe to.
 * @source: The name of the kernel module or subsystem that owns this
 *   subscription.
 * @opts: Whether the topic shall be exported to user space via sysfs.
 *   Pass TOPIC_SYSFS_READ, TOPIC_SYSFS_WRITE, TOPIC_SYSFS_READ_WRITE, or
 *   TOPIC_DEFAULTS. The latter does not expose the topic over sysfs.
 *   Also, supports the options TOPIC_NOTIFY_NOW to indicate that the callback
 *   shall be called immediately with the current value for snapshotting and
 *   TOPIC_NOTIFY_SAME to indicate that the callback shall be called when
 *   a publisher sets the same value again.
 *
 * @example
 * static my_sub = TOPIC_SUBSCRIPTION_INIT(my_sub, "my_topic", "my_module",
 *   TOPIC_SYSFS_READ | TOPIC_NOTIFY_NOW);
 */
#define TOPIC_SUBSCRIPTION_INIT(subscription, name, fn, sink, opts)            \
	{                                                                      \
		.list_node = LIST_HEAD_INIT((subscription).list_node),         \
		.topic_name = (name),                                          \
		.topic = NULL,                                                 \
		.notify_call = (fn),                                           \
		.sink_name = (sink),                                           \
		.options = (opts),                                             \
		.data = NULL                                                   \
	}

/**
 * TOPIC_GLOBAL_SUBSCRIPTION_INIT(subscription, fn, sink, opts) - Initialize a
 * global subscription for all topics.
 * @subscription: Subscription variable to initialize.
 * @source: The name of the kernel module or subsystem that owns this
 *   subscription.
 * @opts: TOPIC_NOTIFY_NOW to indicate that the callback shall be called
 *   immediately once for all existing topics.
 *   TOPIC_NOTIFY_SAME to indicate that the callback shall be called when
 *   a publisher sets the same value again to any topic.
 *
 * @example
 * static my_global_sub = TOPIC_SUBSCRIPTION_INIT(my_global_sub, "my_module",
 *   TOPIC_NOTIFY_NOW);
 */
#define TOPIC_GLOBAL_SUBSCRIPTION_INIT(subscription, fn, sink, opts)           \
	TOPIC_SUBSCRIPTION_INIT(subscription, NULL, fn, sink, opts)

/**
 * DEFINE_TOPIC_SUBSCRIPTION(subscription, name, fn, sink, opts) - Defines a
 * subscription for a particular topic referred-to by its name.
 * @subscription: Subscription variable to define.
 * @name: The name of the topic to subscribe to.
 * @source: The name of the kernel module or subsystem that owns this
 *   subscription.
 * @opts: Whether the topic shall be exported to user space via sysfs.
 *   Pass TOPIC_SYSFS_READ, TOPIC_SYSFS_WRITE, TOPIC_SYSFS_READ_WRITE, or
 *   TOPIC_DEFAULTS. The latter does not expose the topic over sysfs.
 *   Also, supports the options TOPIC_NOTIFY_NOW to indicate that the callback
 *   shall be called immediately with the current value for snapshotting and
 *   TOPIC_NOTIFY_SAME to indicate that the callback shall be called when
 *   a publisher sets the same value again.
 *
 * @example
 * static DEFINE_TOPIC_SUBSCRIPTION(my_sub, "my_topic", "my_module",
 *   TOPIC_SYSFS_READ | TOPIC_NOTIFY_NOW);
 */
#define DEFINE_TOPIC_SUBSCRIPTION(subscription, name, fn, sink, opts)          \
	struct topic_subscription_t subscription =                             \
		TOPIC_SUBSCRIPTION_INIT(subscription, name, fn, sink, opts)

/**
 * TOPIC_GLOBAL_SUBSCRIPTION_INIT(subscription, fn, sink, opts) - Defines a
 * global subscription for all topics.
 * @subscription: Subscription variable to define.
 * @source: The name of the kernel module or subsystem that owns this
 *   subscription.
 * @opts: TOPIC_NOTIFY_NOW to indicate that the callback shall be called
 *   immediately once for all existing topics.
 *   TOPIC_NOTIFY_SAME to indicate that the callback shall be called when
 *   a publisher sets the same value again to any topic.
 *
 * @example
 * static DEFINE_GLOBAL_TOPIC_SUBSCRIPTION(my_global_sub, "my_module",
 *   TOPIC_NOTIFY_NOW);
 */
#define DEFINE_GLOBAL_TOPIC_SUBSCRIPTION(subscription, fn, sink, opts)         \
	struct topic_subscription_t subscription =                             \
		TOPIC_GLOBAL_SUBSCRIPTION_INIT(subscription, fn, sink, opts)

/**
 * subscribe_to_topic(subscription) - Subscribes for topic updates.
 * @subscription: topic_subscriber_t struct, previously defined with
 *   DEFINE_TOPIC_SUBSCRIPTION() or initialized with TOPIC_SUBSCRIPTION_INIT().
 * @value: Optional initial value to set if the topic does not exist yet.
 * @returns 0 if successful.
 *
 * The topic does not have to exist when subscribing. If it does not exist it
 * will be created automatically. After subscribing with the topic, the
 * notify_call function is called whenever the topic's value is updated.
 *
 * @note This function must be called while running in the nmworker thread.
 * If called from outside this worker thread, call synchronize_to_worker()
 * before calling one of the functions and unsynchronize_from_worker() after.
 */
int subscribe_to_topic(struct topic_subscription_t *subscription, const char *value);

/**
 * unsubscribe_from_topic(subscription) - Unsubscribes for topic updates.
 * @subscription: topic_subscriber_t struct, previously subscribed with
 *   subscribe_to_topic().
 *
 * @note This function must be called while running in the nmworker thread.
 * If called from outside this worker thread, call synchronize_to_worker()
 * before calling one of the functions and unsynchronize_from_worker() after.
 */
void unsubscribe_from_topic(struct topic_subscription_t *subscription);

#endif /* _BROKER_H */
