#ifndef WORKER_H_
#define WORKER_H_

#include <linux/init.h>
#include <linux/types.h>
#include <linux/kthread.h>
#include <linux/semaphore.h>

/**
 * Instead of using pure mutexes or other locks, the broker is protected by
 * running all calls in a separate worker thread, "nmworker".
 *
 * Why not just traditional locks?
 *
 * For each topic, the broker contains a notification chain, i.e. a linked list
 * that has to be traversed and, for each entry, a callback function has to be
 * called. This notification chain must be protected from concurrent thread
 * access, e.g., when two user-space processes write to the same sysfs file.
 * However, keeping a lock while executing a callback is dangerous, because
 * the callback could itself access the topic, leading to nested locking, which
 * is allowed only for a few lock types. Also, virtually every shared resource
 * has to be protected: the code ends up sprinkled with locks.
 *
 * What is the alternative?
 *
 * Instead, we define that all broker code and the majority of code of the
 * dependent drivers (sysstate, extmod, etc.) have to run in the context of
 * the "nmworker" thread - a kernel worker thread. Timers can be realized
 * easily by using deferred work. But other calls to the broker originate from
 * other thredas (e.g. user-space processes). It would be cumbersome to convert
 * all access to work jobs that are executed asynchronously, especially if an
 * access expects a return value. For this case, we implemented the
 * synchronization mechanism in this file. You are safe if you call
 * synchronize_to_worker() before and  unsynchronize_from_worker() after
 * accessing a critical section from any non-atomic context, e.g.,
 *
 * struct topic_t *my_topic;
 * if (synchronize_to_woker()) {
 *   my_topic = get_topic("my-topic");
 *   // do something with my_topic
 *   unsynchronize_from_worker();
 * }
 *
 * How does synchronization work?
 *
 * Instead of really executing work (e.g. get-topic() above) in the context of
 * the nmworker thread, synchronize_to_worker() stops the worker between two
 * work jobs and unsynchronize_from_work() releases the nmworker thread. Hence,
 * actually the critical section is accessed from the original context but
 * we are sure that no other thread - not even nmworker - is accessing critical
 * sections at the same time because the worker thread is suspended.
 *
 * Accessing Thread     Synchronizer        nmworker           Critical Section
 * ================     ============        ========           ================
 *    |                      |                 |                          |
 * 1  |----synchronize()---->|                 |                          |
 * 2  |               take sem_entry           |                          |
 * 3  |                      |---\ work        |                          |
 * 4  |   take sem_wait_for_sync |             |                          |
 *    |                      .   |             |                          |
 * 5  |                      .   \--handle()-->|                          |
 * 6  |                      .<--give sem_wait_for_sync                   |
 * 7  |                      |   take sem_wait_for_unsync                 |
 * 8  |< - - - - - - - - - - |                 .                          |
 *    |                      |                 .                          |
 * 9  |----------------------------------------------- access ----------->|
 *    |                      |                 .                          |
 * 10 |---unsynchronize()--->|                 .                          |
 * 11 |             give sem_wait_for_unsync-->.                          |
 * 12 |                give sem_entry          |                          |
 * 13 |< - - - - - - - - - - |                 |                          |
 *
 * 1.  Some context (e.g., sysfs write) wants to access a critical resource
 *     (e.g., change the value of a topic). Before accessing the resource
 *     at step 8, the accesing thread calls synchronize_to_worker().
 * 2.  The syncrhonize_to_worker() function first takes the entry semaphore.
 *     This restircts only one thread to be allowed to enter the synrhronization
 *     procedure.
 * 3.  Then, the synchronize_to_worker() function schedules an asynchronous work
 *     job with the nmworker thread.
 * 4.  After scheduling the work, the synchronize_to_worker() function blocks on
 *     the sync_sem_wait_for_sync semaphore, waiting for the worker thread to
 *     acknowledge its suspension.
 * 5.  The work is eventually handled by the nmworker thread by the function
 *     handle_sync_work().
 * 6.  The handler gives the semaphore the accessing thread is waiting for.
 * 7.  ...and immediately takes another semaphore: sync_sem_wait_for_sync. Now
 *     the nmworker thread is blocked in the synchronization handler. We are
 *     sure that no one accesses the critical section.
 * 8.  The synchronize_to_worker() call returns.
 * 9.  The accessing thread is now safe to access the critical section from its
 *     own thread. This comes with the big advantage that all function calls
 *     made while synchronized, can return values, which would not be possible
 *     easily when using asynchronous worker jobs.
 * 10. When the accesing thread has completed its calls to the critical sections
 *     it calls unsynchronize_from_worker(),
 * 11. ...which give the sync_sem_wait_for_unsync semaphore the synchronization
 *     handler in nmworker is waiting for. This releases the nmworker thread,
 *     which continues executing other scheduled work, e.g., timers.
 * 12. Finally the unsynchronize_from_worker() function gives the sync_sem_entry
 *     semaphore, allowing the next accessing thread to enter the
 *     synchronziation procedure.
 */

extern struct kthread_worker *nm_worker;

int __init init_worker(void);
void __exit cleanup_worker(void);

/**
 * synchronize_to_worker() - Synchronizes any non-atomic context to the
 * nmworker thread before accessing a critical section.
 * Returns 1 if synchronization was successful.
 *
 * Note: After accessing the the critical section, unsynchronize_from_worker()
 * must be called. The two function must always be called in symmetry.
 */
bool synchronize_to_worker(void);

/**
 * unsynchronize_from_worker() - Unsynchronizes from the nmworker thread after
 * accessing the critical section. Must be called after every successful
 * synchronize_to_worker() call.
 */
void unsynchronize_from_worker(void);

#endif /* WORKER_H_ */
